"""Provide Ant`s Request, Response, Item and Extractor."""
from typing import Any, Optional, Tuple, Type, Union, List, \
    DefaultDict, AnyStr, IO, Callable, Generator, TypeVar
from collections import defaultdict
import logging
import re

from aiohttp import ClientResponse, ClientRequest
from aiohttp.client import DEFAULT_TIMEOUT
from lxml import html
import jpath
import ujson

from .exceptions import ItemExtractError


class Request(ClientRequest):
    def __init__(self, *args, timeout: float = DEFAULT_TIMEOUT.total,
                 response_in_stream: bool = False, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.response_in_stream = response_in_stream
        self.timeout = timeout
        # store data obj
        self.data: Optional[Union[AnyStr, dict, IO]] = kwargs.get('data', None)


class Response(ClientResponse):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._text = None
        self._html_element = None
        self._json = None

    def get_text(self, encoding: Optional[str] = None,
                 errors: str = 'strict') -> str:

        if self._body is None:
            raise ValueError('Read stream first')
        if self._text is None:
            if encoding is None:
                encoding = self.get_encoding()
            self._text = self._body.decode(encoding, errors=errors)
        return self._text

    @property
    def simple_text(self) -> str:
        return self.get_text(errors='ignore')

    def get_json(self, loads: Callable = ujson.loads):
        if self._json is None:
            self._json = loads(self.simple_text)
        return self._json

    @property
    def simple_json(self) -> Any:
        return self.get_json()

    @property
    def html_element(self) -> html.HtmlElement:
        if self._html_element is None:
            self._html_element = html.fromstring(self.simple_text)
        return self._html_element


class CustomNoneType:
    """Different with "None" obj ("null" in json)
    """
    pass


Item = TypeVar('Item')
Things = Union[Request, Response, Item]


class ItemExtractor:
    extract_with_take_first = 'take_first'
    extract_with_join_all = 'join_all'
    extract_with_do_nothing = 'do_nothing'

    def __init__(self, item_class: Type[Item]) -> None:
        self.item_class = item_class
        self.logger = logging.getLogger(self.__class__.__name__)
        self.paths: DefaultDict[str, List[Tuple[str, str, str]]] = \
            defaultdict(list)

    def add_xpath(self, key: str, xpath: str,
                  extract_type=extract_with_take_first):
        self.paths[key].append(('xpath', xpath, extract_type))

    def add_regex(self, key: str, pattern: str,
                  extract_type=extract_with_take_first):
        self.paths[key].append(('regex', pattern, extract_type))

    def add_jpath(self, key: str, jpath, extract_type=extract_with_take_first):
        self.paths[key].append(('jpath', jpath, extract_type))

    @staticmethod
    def _regex_search(pattern: str, data: Any) -> List[str]:
        # convert data to string
        if isinstance(data, Response):
            data = data.simple_text
        elif isinstance(data, html.HtmlElement):
            data = html.tostring(data, encoding='unicode')
        else:
            data = str(data)
        return re.findall(pattern, data)

    @staticmethod
    def _jpath_search(pattern: str, data: Any) -> List[Any]:
        # convert data to json object
        if isinstance(data, Response):
            data = data.simple_json
        elif isinstance(data, str):
            data = ujson.loads(data)
        return jpath.get_all(pattern, data)

    @staticmethod
    def _xpath_search(pattern: str, data: Any
                      ) -> List[Union[str, html.HtmlElement]]:
        if isinstance(data, Response):
            data = data.html_element
        elif not isinstance(data, html.HtmlElement):
            data = html.fromstring(str(data))
        return data.xpath(pattern)

    @staticmethod
    def extract_value(_type: str, pattern: str, data: Any,
                      extract_type=extract_with_take_first) -> Any:
        if _type == 'xpath':
            extract_value = ItemExtractor._xpath_search(pattern, data)
        elif _type == 'regex':
            extract_value = ItemExtractor._regex_search(pattern, data)
        elif _type == 'jpath':
            extract_value = ItemExtractor._jpath_search(pattern, data)
        else:
            raise ValueError('The type: {:s} not support'.format(_type))
        # handle by extract type
        if extract_type == ItemExtractor.extract_with_take_first:
            extract_value = extract_value[0]
        elif extract_type == ItemExtractor.extract_with_join_all:
            extract_value = list(filter(lambda x: isinstance(x, str),
                                        extract_value))  # join string only
            extract_value = ''.join(extract_value)
        return extract_value

    def extract(self, data: Any) -> Item:
        """Extract item from response or other data by xpath,
        jpath or regex."""
        self.logger.debug('Extract item: {:s} with path: {:s}'.format(
            self.item_class.__name__, str(self.paths)))
        item = self.item_class()
        for key, paths in self.paths.items():
            value = CustomNoneType()
            for path_type, path, extract_type in paths:
                try:
                    extract_value = ItemExtractor.extract_value(
                        path_type, path, data, extract_type=extract_type)
                except IndexError:  # IndexError is often raised
                    continue
                # check multiple path`s result
                if not isinstance(value,
                                  CustomNoneType) and value != extract_value:
                    raise ItemExtractError(
                        'Match different result: {:s} and {:s} for paths: '
                        '{:s}'.format(value, extract_value, str(paths)))
                value = extract_value
            if not isinstance(value, CustomNoneType):
                setattr(item, key, value)
        return item


class ItemNestExtractor(ItemExtractor):
    def __init__(self, root_path_type: str, root_path: str,
                 item_class: Type[Item]) -> None:
        self.root_path_type = root_path_type
        self.root_path = root_path
        super().__init__(item_class)

    def extract(self, response: Response):
        raise NotImplementedError('This method is dropped in this class')

    def extract_items(self, response: Response
                      ) -> Generator[Item, None, None]:
        base_data = self.extract_value(
            self.root_path_type, self.root_path, response,
            extract_type=self.extract_with_do_nothing)
        for data in base_data:
            yield super().extract(data)


__all__ = ['Request', 'Response', 'ItemExtractor', 'ItemNestExtractor',
           'Things'] + \
          [var for var in vars().keys() if 'Field' in var]



import asyncio
import os

import pytest
from tenacity import RetryError
import asyncio
from yarl import URL

from ant_nest import *
from .test_things import fake_response


@pytest.mark.asyncio
async def test_ant():
    class TestPipeline(Pipeline):
        def __init__(self):
            super().__init__()
            self.count = 0

        def process(self, thing):
            self.count += 1
            return thing

    class TestAnt(Ant):
        item_pipelines = [TestPipeline()]
        request_retries = 0

        async def run(self):
            await self.request('http://test.com')
            await self.collect(object())
            assert self.item_pipelines[0].count == 1

        async def _request(self, req: Request):
            return fake_response(b'')

    ant = TestAnt()
    assert ant.name == 'TestAnt'
    await ant.main()

    class Test2Ant(TestAnt):
        async def run(self):
            raise Exception('Test exception')

    await Test2Ant().main()


@pytest.mark.asyncio
async def test_ant_with_retry():
    class Test2Ant(Ant):
        request_retries = 2
        request_retry_delay = 0.1

        def __init__(self):
            super().__init__()
            self.min_retries = 2
            self.retries = 0

        async def run(self):
            return None

        async def _request(self, req: Request):
            if self.retries < self.min_retries:
                self.retries += 1
                raise IOError()
            else:
                res = fake_response(b'')
                res.status = 200
                return res

    await Test2Ant().request('https://www.test.com')
    ant = Test2Ant()
    ant.request_retries = 1
    with pytest.raises(RetryError):
        await ant.request('https://www.test.com')
    # with params
    ant.retries = 0
    ant.request_retries = 3
    await ant.request('https://www.test.com')
    ant.retries = 0
    with pytest.raises(IOError):
        await ant.request('https://www.test.com', retries=0)

    await ant.main()
    # retry condition

    class Test3Ant(Test2Ant):
        async def _request(self, req: Request):
            if self.retries < self.min_retries:
                self.retries += 1
                res = fake_response(b'')
                res.status = 500
                return res
            else:
                res = fake_response(b'')
                res.status = 200
                return res

    await Test3Ant().request('https://www.test.com')
    ant = Test3Ant()
    ant.request_retries = 1
    with pytest.raises(RetryError):
        await ant.request('https://www.test.com')

    await ant.main()


@pytest.mark.asyncio
async def test_pipelines():

    class TestPipeline(Pipeline):
        async def process(self, thing):
            raise TypeError('Test error')

        def on_spider_open(self):
            raise TypeError('Test error')

        def on_spider_close(self):
            raise TypeError('Test error')

    class TestAnt(Ant):
        item_pipelines = [TestPipeline()]

        async def run(self):
            return None

    pls = [Pipeline() for x in range(10)]
    thing = object()
    ant = TestAnt()
    assert thing is await ant._handle_thing_with_pipelines(thing, pls)
    # with exception
    pls[5] = TestPipeline()
    with pytest.raises(TypeError):
        await ant._handle_thing_with_pipelines(thing, pls)
    # exception will be ignored in "open" and "close" method
    with pytest.raises(TypeError):
        await ant.open()
    with pytest.raises(TypeError):
        await ant.close()


@pytest.mark.asyncio
async def test_ant_report():

    class FakePipeline(Pipeline):
        def process(self, thing: Request):
            if thing.host == 'error':
                raise ThingDropped
            return thing

    class FakeAnt(Ant):
        request_pipelines = [FakePipeline()]
        request_retries = 0

        async def run(self):
            pass

        async def _request(self, req: Request):
            return fake_response(b'')

    ant = FakeAnt()
    # request report
    await ant.request('http://test.com')
    assert ant._reports['Request'][1] == 1
    assert ant._reports['Request'][0] == 0
    with pytest.raises(ThingDropped):
        await ant.request('http://error')
    assert ant._drop_reports['Request'][1] == 1

    ant._last_time -= ant._last_time + 1  # report
    ant.report(Request('GET', URL('http://test')))
    assert ant._reports['Request'][0] == 1
    ant.report(Request('GET', URL('http://test')), dropped=True)
    assert ant._drop_reports['Request'][0] == 1

    await ant.main()


@pytest.mark.asyncio
async def test_with_real_request():
    httpbin_base_url = os.getenv('TEST_HTTPBIN', 'http://localhost:8080/')

    class TestAnt(Ant):
        async def run(self):
            return None

    ant = TestAnt()
    res = await ant.request(httpbin_base_url)
    assert res.status == 200
    # method
    for method in ['GET', 'POST', 'DELETE', 'PUT', 'PATCH', 'HEAD']:
        res = await ant.request(httpbin_base_url + 'anything', method=method)
        assert res.status == 200
        if method != 'HEAD':
            assert res.simple_json['method'] == method
        else:
            assert res.simple_text == ''
        # short way
        res = await getattr(ant, method.lower())(httpbin_base_url + 'anything')
        assert res.status == 200
        if method != 'HEAD':
            assert res.simple_json['method'] == method
        else:
            assert res.simple_text == ''
    # params
    res = await ant.request(httpbin_base_url + 'get?k1=v1&k2=v2')
    assert res.status == 200
    assert res.simple_json['args']['k1'] == 'v1'
    assert res.simple_json['args']['k2'] == 'v2'
    res = await ant.get(httpbin_base_url + 'get', params={'k1': 'v1', 'k2': 'v2'})
    assert res.status == 200
    assert res.simple_json['args']['k1'] == 'v1'
    assert res.simple_json['args']['k2'] == 'v2'
    # data with str
    res = await ant.post(httpbin_base_url + 'post', data='Test data')
    assert res.status == 200
    assert res.simple_json['data'] == 'Test data'
    # data with dict
    res = await ant.request(httpbin_base_url + 'post', method='POST', data={'k1': 'v1'})
    assert res.status == 200
    assert res.simple_json['form']['k1'] == 'v1'
    # data with bytes
    res = await ant.request(httpbin_base_url + 'post', method='POST', data=b'12345')
    assert res.status == 200
    assert res.simple_json['data'] == '12345'
    # data with file
    with open('tests/test.html', 'r') as f:
        file_content = f.read()
        f.seek(0)
        res = await ant.request(httpbin_base_url + 'post', method='POST', data=f)
        assert res.status == 200
        assert res.simple_json['data'] == file_content
    # headers
    res = await ant.request(httpbin_base_url + 'headers', headers={'Custom-Key': 'test'})
    assert res.status == 200
    assert res.headers['Content-Type'] == 'application/json'
    assert res.simple_json['headers']['Custom-Key'] == 'test'
    # cookies
    res = await ant.request(httpbin_base_url + 'cookies', cookies={'Custom-Key': 'test'})
    assert res.status == 200
    assert res.simple_json['cookies']['Custom-Key'] == 'test'
    # get cookies
    ant.request_allow_redirects = False
    res = await ant.request(httpbin_base_url + 'cookies/set?k1=v1&k2=v2')
    assert res.status == 302
    assert res.cookies['k1'].value == 'v1'
    assert res.cookies['k2'].value == 'v2'
    # redirects
    ant.request_allow_redirects = True
    ant.request_max_redirects = 3
    res = await ant.request(httpbin_base_url + 'redirect/2')
    assert res.status == 200
    # with http proxy
    proxy = os.getenv('TEST_HTTP_PROXY', 'http://bugman:letmein@localhost:3128')
    ant.request_proxies.append(proxy)
    res = await ant.request('http://httpbin.org/anything')
    assert res.status == 200
    # no proxy anymore
    ant.request_proxies.pop()
    res = await ant.request('http://httpbin.org/anything')
    assert res.status == 200
    # set proxy by request
    res = await ant.get('http://httpbin.org/anything', proxy=proxy)
    assert res.status == 200
    # with stream
    ant.response_in_stream = True
    res = await ant.request(httpbin_base_url + 'anything')
    assert res.status == 200
    with pytest.raises(ValueError):
        res.simple_text
    while True:
        chunk = await res.content.read(10)
        if len(chunk) == 0:
            break
    # set streaming by request
    res = await ant.request(httpbin_base_url + 'anything',
                            response_in_stream=False)
    assert res.status == 200
    assert res.simple_text is not None

    await ant.close()



import os
import time
from datetime import datetime
import io
from unittest import mock

import pytest
from yarl import URL
import aiofiles

from ant_nest import *
from .test_things import fake_response


@pytest.mark.asyncio
async def test_pipeline():
    pl = Pipeline()
    await pl.process(Request('GET', URL('https://test.com')))


def test_response_filter_error_pipeline():
    pl = ResponseFilterErrorPipeline()
    res = fake_response(b'')
    err_res = fake_response(b'')
    res.status = 200
    err_res.status = 403
    assert res is pl.process(res)
    with pytest.raises(ThingDropped):
        pl.process(err_res)


def test_request_duplicate_filter_pipeline():
    pl = RequestDuplicateFilterPipeline()
    req = Request('GET', URL('http://test.com'))
    assert pl.process(req) is req
    with pytest.raises(ThingDropped):
        pl.process(req)


def test_item_print_pipeline(item_cls):
    pl = ItemPrintPipeline()
    item = item_cls()
    item.count = 3
    item.info = 'hi'
    assert pl.process(item) is item


def test_item_filed_replace_pipeline(item_cls):
    pl = ItemFieldReplacePipeline(['info'])
    item = item_cls()
    item.info = 'hi\n,\t\r ant\n'
    pl.process(item)
    assert item.info == 'hi, ant'


@pytest.mark.asyncio
async def test_item_base_file_dump_pipeline():
    pl = ItemBaseFileDumpPipeline()
    await pl.dump('/dev/null', 'Hello World')
    await pl.dump('/dev/null', b'Hello World')
    await pl.dump('/dev/null', io.StringIO('Hello World'))
    await pl.dump('/dev/null', io.BytesIO(b'Hello World'))
    await pl.dump('/dev/null', open('./tests/test.html'), buffer_size=4)
    async with aiofiles.open('./tests/test.html') as f:
        await pl.dump('/dev/null', f)
    async with aiofiles.open('./tests/test.html', 'rb') as f:
        await pl.dump('/dev/null', f, buffer_size=4)

    with pytest.raises(ValueError):
        await pl.dump('/dev/null', None)


@pytest.mark.asyncio
async def test_item_json_dump_pipeline(item_cls):
    pl = ItemJsonDumpPipeline(to_dict=lambda x: x)
    item = item_cls()
    item.count = 1
    assert pl.process(item) is item
    item = item_cls()
    item.info = 'hi'
    pl.process(item)
    await pl.on_spider_close()

    # clean file
    ci = os.getenv('TEST_HOST', 'localhost')
    if ci == 'localhost':
        os.remove('./Item.json')


def test_request_user_agent_pipeline():
    pl = RequestUserAgentPipeline(user_agent='ant')
    req = Request('GET', URL('https://www.hi.com'))
    assert pl.process(req) is req
    assert req.headers['User-Agent'] == 'ant'

    req.headers['User-Agent'] = 'custom'
    assert pl.process(req).headers['User-Agent'] == 'custom'


def test_request_random_user_agent_pipeline():
    pl = RequestRandomUserAgentPipeline()
    req = Request('GET', URL('https://www.hi.com'))
    assert pl.process(req) is req
    assert req.headers.get('User-Agent') is not None

    req.headers['User-Agent'] = 'custom'
    assert pl.process(req).headers['User-Agent'] == 'custom'

    with pytest.raises(ValueError):
        RequestRandomUserAgentPipeline(system='something')

    with pytest.raises(ValueError):
        RequestRandomUserAgentPipeline(browser='something')

    pl = RequestRandomUserAgentPipeline(system='UnixLike', browser='Firefox')
    user_agent = pl.create()
    assert 'X11' in user_agent
    assert 'Firefox' in user_agent


@pytest.mark.asyncio
async def test_item_email_pipeline(item_cls):
    class FakeSMTP:
        async def send_message(self, msg):
            pass

        async def connect(self, *args, **kwargs):
            pass

        async def starttls(self):
            pass

        async def login(self, *args, **kwargs):
            pass

        def close(self):
            pass

    pl = ItemEmailPipeline('test', 'a@b.c', 'letmein', 'localhost', 25,
                           recipients=['b@a.c', 'c@b.a'])

    with mock.patch('aiosmtplib.SMTP', new=FakeSMTP):
        fsmtp = await pl.create_smtp()
        pl.starttls = True
        fsmtp = await pl.create_smtp()

        await pl.send(fsmtp, 'test', 'test')
        with open('./tests/test.html') as f:
            await pl.send(fsmtp, 'test', 'test', attachments=[f])

        # with item
        item = item_cls()
        item.info = 'hi'
        pl.process(item)
        await pl.on_spider_close()


@pytest.mark.asyncio
async def test_item_mysql_pipeline():
    mysql_server = os.getenv('TEST_MYSQL_SERVER', 'localhost')
    mysql_port = int(os.getenv('TEST_MYSQL_PORT', 3306))
    mysql_user = os.getenv('TEST_MYSQL_USER', 'root')
    mysql_password = os.getenv('TEST_MYSQL_PASSWORD', 'letmein')

    bpl = ItemBaseMysqlPipeline(host=mysql_server, port=mysql_port,
                                user=mysql_user, password=mysql_password,
                                database='mysql', table='',
                                to_dict=lambda x: x,
                                buffer_length=1)
    await bpl.on_spider_open()
    await bpl.push_data('''DROP DATABASE IF EXISTS test;
                           CREATE DATABASE test;''')
    await bpl.push_data(
        '''CREATE TABLE test.test (
        `id` int(11) unsigned NOT NULL AUTO_INCREMENT,
        `test` TEXT DEFAULT NULL,
        `test_bool` BOOL DEFAULT NULL,
        `test_int` INT DEFAULT NULL,
        `test_float` FLOAT DEFAULT NULL,
        `test_bytes` BLOB DEFAULT NULL,
        `test_datetime` DATETIME DEFAULT NULL,
        PRIMARY KEY (`id`)
        ) ENGINE=InnoDB AUTO_INCREMENT=3 DEFAULT CHARSET=utf8;''')

    test_item = dict(test='I ant', test_bool=False, test_int=1, test_float=0.3,
                     test_bytes=b'\xf0\x9f\x91\x8d',
                     test_datetime=datetime.now())
    ibpl = ItemMysqlInsertPipeline(host=mysql_server, port=mysql_port,
                                   user=mysql_user, password=mysql_password,
                                   database='test', table='test',
                                   to_dict=lambda x: x,
                                   buffer_length=1)
    await ibpl.on_spider_open()
    assert test_item is await ibpl.process(test_item)
    data = await ibpl.pull_data('SELECT * FROM test')
    assert test_item['test'] == data[0]['test']

    ubpl = ItemMysqlUpdatePipeline(host=mysql_server, port=mysql_port,
                                   user=mysql_user, password=mysql_password,
                                   database='test', table='test',
                                   to_dict=lambda x: x,
                                   primary_key='id', buffer_length=1)
    await ubpl.on_spider_open()
    test_item['id'] = data[0]['id']
    test_item['test'] = 'I ANT'
    assert test_item is await ubpl.process(test_item)
    data = await ubpl.pull_data('SELECT * FROM test')
    assert test_item['test'] == data[0]['test']
    test_item['test'] = None
    assert test_item is await ubpl.process(test_item)
    data = await ubpl.pull_data('SELECT * FROM test')
    assert test_item['test'] == data[0]['test']

    iubpl = ItemMysqlInsertUpdatePipeline(
        update_keys=['test'],
        host=mysql_server, port=mysql_port, user=mysql_user,
        password=mysql_password,
        to_dict=lambda x: x,
        database='test', table='test', buffer_length=1)
    await iubpl.on_spider_open()
    test_item['test'] = 'I love ant!'
    assert test_item is await iubpl.process(test_item)
    data = await iubpl.pull_data('SELECT * FROM test')
    assert test_item['test'] == data[0]['test']

    await ubpl.on_spider_close()
    await ibpl.on_spider_close()
    await iubpl.on_spider_close()
    await bpl.push_data('DROP TABLE test.test;DROP DATABASE test')


@pytest.mark.asyncio
async def test_redis_pipeline():
    redis_address = os.getenv('TEST_REDIS_ADDRESS', 'redis://localhost:6379')
    pl = ItemBaseRedisPipeline(redis_address)
    pool = await pl.create_redis()
    with await pool as conn:
        value = 'value'
        await conn.set('key', value)
        assert await conn.get('key') == value
    pool.close()
    await pool.wait_closed()



from typing import Optional, List, Tuple, DefaultDict, Dict, Any, IO, Union, \
    Sequence, AnyStr, Callable
import asyncio
import logging
from collections import defaultdict
import ujson
import os
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
import random
import re

import aiomysql
import aioredis
import aiosmtplib
import aiofiles
from aiohttp.http import SERVER_SOFTWARE
from aiohttp import hdrs

from .things import Things, Response, Request, Item
from .exceptions import ThingDropped


class Pipeline:
    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    async def on_spider_open(self) -> None:
        """Call when ant open, this method can be coroutine function"""

    async def on_spider_close(self) -> None:
        """Call when ant close, this method can be coroutine function"""

    async def process(self, thing: Things) -> Things:
        """Process things, this method can be coroutine function
        Raise ThingDropped when drop one thing

        :raise ThingDropped
        """
        return thing


# Response pipelines
class ResponseFilterErrorPipeline(Pipeline):
    def process(self, thing: Response) -> Union[Things, Exception]:
        if thing.status >= 400:
            raise ThingDropped('Respose status is {:d}'.format(thing.status))
        else:
            return thing


# Request pipelines
class RequestDuplicateFilterPipeline(Pipeline):
    def __init__(self):
        self.__request_urls = set()
        super().__init__()

    def process(self, thing: Request) -> Union[Things, Exception]:
        if thing.url in self.__request_urls:
            raise ThingDropped('Request duplicate!')
        else:
            self.__request_urls.add(thing.url)
            return thing


class RequestUserAgentPipeline(Pipeline):
    user_agent = 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 ' \
                 '(KHTML, like Gecko) Chrome/51.0.2704.103 Safari/537.36'

    def __init__(self, user_agent=user_agent):
        super().__init__()
        self.user_agent = user_agent

    def process(self, thing: Request) -> Request:
        if thing.headers.get(hdrs.USER_AGENT) == SERVER_SOFTWARE:
            thing.headers[hdrs.USER_AGENT] = self.user_agent
        return thing


class RequestRandomUserAgentPipeline(Pipeline):
    """Create simple and common user agent for request by random,
    It`s easy to add new rule.
    """
    USER_AGENT_FORMAT = 'Mozilla/5.0 ({system}) {browser}'
    SYSTEM_FORMATS = {
        'UnixLike': 'X11; {unix-like_os} {cpu_type}',
        'MacOS': 'Macintosh; Intel Mac OS X {macos_version}',
        'Windows': 'Windows NT {windows_version}',
        'Android': 'Android {android_version}; Linux',
        'iOS': '{ios_driver}; CPU OS {ios_version} like Mac OS X',
    }
    BROWSER_FORMATS = {
        'Firefox': 'Gecko/20100101 Firefox/{firefox_version}',
        'Safari': 'AppleWebKit/{webkit_version} (KHTML, like Gecko) '
                  'Version/{safari_version} Safari/{safari_version2}',
        'Chrome': 'AppleWebKit/{webkit_version} (KHTML, like Gecko) '
                  'Chrome/{chrome_version} Safari/{safari_version2}',
    }
    FORMAT_VARS = {
        'unix-like_os': ('Linux', 'FreeBSD'),
        'cpu_type': ('x86_64', 'i386', 'amd64'),
        'macos_version': (
            '10_10_1', '10_11_1', '10_11_2', '10_12_2', '10_12_3', '10_13_3'),
        'windows_version': ('5_0', '5_1', '5_2', '6_0', '6_1', '10_0'),
        'android_version': (
            '4_4', '5_0', '5_1', '6_0', '6_1', '7_0', '7_1', '8_0'),
        'ios_driver': ('iPone', 'iPod', 'iPad'),
        'ios_version': (
            '6_0', '7_0', '8_1', '9_2', '10_3', '10_2_3', '11_3_3'),
        'firefox_version': ('27.3', '28.0', '31.0', '40.1'),
        'webkit_version': ('533.18.1', '533.19.4', '533.20.25', '534.55.3'),
        'safari_version': ('4.0.1', '5.0.4', '5.1.3', '6.0', '7.0.3'),
        'safari_version2': ('530.19.1', '531.9', '533.16', '533.20.27'),
        'chrome_version': (
            '41.0.2226.0', '60.0.1325.223', '62.0.1532.123', '64.0.3282.119')
    }

    def __init__(self, system: str = 'random', browser: str = 'random'):
        if system != 'random' and system not in self.SYSTEM_FORMATS.keys():
            raise ValueError(
                'The system {:s} is not supported!'.format(system))
        if browser != 'random' and browser not in self.BROWSER_FORMATS.keys():
            raise ValueError(
                'The browser {:s} is not supported!'.format(browser))

        self.system = system
        self.browser = browser
        super().__init__()

    @staticmethod
    def choice(data: Sequence[str]) -> str:
        return random.choice(data)

    def _format(self, pattern: str) -> str:
        """format system or browser pattern string"""
        kv = {}
        keys = re.findall('{(\S+?)}', pattern)

        for key in keys:
            kv[key] = self.choice(self.FORMAT_VARS[key])

        return pattern.format(**kv)

    def create(self) -> str:
        if self.system != 'random':
            system_format = self.SYSTEM_FORMATS[self.system]
        else:
            system_format = self.SYSTEM_FORMATS[
                self.choice(list(self.SYSTEM_FORMATS.keys()))]

        if self.browser != 'random':
            browser_format = self.BROWSER_FORMATS[self.browser]
        else:
            browser_format = self.BROWSER_FORMATS[
                self.choice(list(self.BROWSER_FORMATS.keys()))]

        return self.USER_AGENT_FORMAT.format(
            system=self._format(system_format),
            browser=self._format(browser_format))

    def process(self, thing: Request) -> Request:
        if thing.headers.get(hdrs.USER_AGENT) == SERVER_SOFTWARE:
            thing.headers[hdrs.USER_AGENT] = self.create()
        return thing


class RequestRandomComputerUserAgentPipeline(Pipeline):
    SYSTEM_FORMATS = {
        'UnixLike': 'X11; {unix-like_os} {cpu_type}',
        'MacOS': 'Macintosh; Intel Mac OS X {macos_version}',
        'Windows': 'Windows NT {windows_version}',
    }


class RequestRandomMobileUserAgentPipeline(Pipeline):
    SYSTEM_FORMATS = {
        'Android': 'Android {android_version}; Linux',
        'iOS': '{ios_driver}; CPU OS {ios_version} like Mac OS X',
    }


# Item pipelines
class ItemPrintPipeline(Pipeline):
    def process(self, thing: Item) -> Item:
        self.logger.info(thing.__repr__())
        return thing


class ItemFieldReplacePipeline(Pipeline):
    def __init__(self, fields: List[str],
                 excess_chars: Tuple[str] = ('\r', '\n', '\t')):
        self.fields = fields
        self.excess_chars = excess_chars
        super().__init__()

    def process(self, thing: Item) -> Item:
        for field in self.fields:
            for char in self.excess_chars:
                if isinstance(getattr(thing, field), str):
                    setattr(
                        thing, field, getattr(thing, field).replace(char, ''))
        return thing


class ItemBaseFileDumpPipeline(Pipeline):

    @classmethod
    async def dump(cls, file_path: str, data: Union[AnyStr, IO],
                   buffer_size: int = 1024 * 1024) -> None:
        """Dump data(binary or text, stream or normal, async or not) to disk file.
        IO data will be closed.
        """
        chunk: Optional[AnyStr] = None
        if isinstance(data, str):
            file_mode = 'w'
        elif isinstance(data, bytes):
            file_mode = 'wb'
        elif hasattr(data, 'read'):  # readable
            chunk = data.read(buffer_size)
            if asyncio.iscoroutine(chunk):
                chunk = await chunk

            if isinstance(chunk, str):
                file_mode = 'w'
            else:
                file_mode = 'wb'
        else:
            raise ValueError('The type {:s} is not supported'.format(
                type(data).__class__.__name__))

        async with aiofiles.open(file_path, file_mode) as file:
            if chunk is not None:  # in streaming
                await file.write(chunk)
                while True:

                    chunk = data.read(buffer_size)
                    if asyncio.iscoroutine(chunk):
                        chunk = await chunk

                    if len(chunk) == 0:
                        break
                    else:
                        await file.write(chunk)

                result = data.close()
                if asyncio.iscoroutine(result):
                    await result
            else:
                await file.write(data)


class ItemJsonDumpPipeline(ItemBaseFileDumpPipeline):
    """Dump item to json during pipeline closing"""

    def __init__(
            self, *, to_dict: Callable[[Item], Dict], file_dir: str = '.'):
        super().__init__()
        self.file_dir = file_dir
        self.data: DefaultDict[str, List[Dict]] = defaultdict(list)
        self.to_dict = to_dict

    def process(self, thing: Item) -> Item:
        self.data[thing.__class__.__name__].append(self.to_dict(thing))
        return thing

    async def on_spider_close(self) -> None:
        for file_name, data in self.data.items():
            data = ujson.dumps(data)
            await self.dump(os.path.join(self.file_dir, file_name + '.json'),
                            data)


class ItemBaseMysqlPipeline(Pipeline):
    def __init__(self, *, host: str, port: int, user: str, password: str,
                 database: str, table: str, to_dict: Callable[[Item], Dict],
                 charset: str = 'utf8', buffer_length: int = 100):
        super().__init__()
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        self.table = table
        self.to_dict = to_dict
        self.charset = charset
        # writing buffer
        self.buffer: List[str] = []
        self.buffer_length = buffer_length
        self.pool: Optional[aiomysql.Pool] = None

    async def on_spider_open(self) -> None:
        self.pool = await self.create_pool()

    async def on_spider_close(self):
        if len(self.buffer) > 0:
            self.buffer_length = 1
            await self.push_data('')
        self.pool.close()
        await self.pool.wait_closed()

    async def create_pool(self) -> aiomysql.Pool:
        return await aiomysql.create_pool(host=self.host, port=self.port,
                                          user=self.user,
                                          password=self.password,
                                          db=self.database,
                                          charset=self.charset,
                                          use_unicode=True)

    async def push_data(self, sql: str) -> None:
        """Run SQL without pulling data like "INSERT" and "UPDATE" command"""
        if self.pool is None:
            raise RuntimeError('Need connection pool!')

        self.logger.debug('Executing SQL: ' + sql)
        self.buffer.append(sql)
        if len(self.buffer) >= self.buffer_length:
            async with self.pool.acquire() as conn:
                async with conn.cursor() as cur:
                    await cur.execute(';'.join(self.buffer))
                await conn.commit()
            self.buffer.clear()

    async def pull_data(self, sql: str) -> Tuple[Dict[str, Any]]:
        """Run SQL with pulling data like "SELECT" command"""
        if self.pool is None:
            raise RuntimeError('Need connection pool!')

        self.logger.debug('Executing SQL: ' + sql)
        async with self.pool.acquire() as conn:
            async with conn.cursor(aiomysql.DictCursor) as cur:
                await cur.execute(sql)
                return await cur.fetchall()

    @staticmethod
    def convert_item_value(value: Any) -> str:
        """Parse value to str for making sql string, eg: False -> '0'"""
        if isinstance(value, bool):
            return '1' if value else '0'
        elif isinstance(value, int) or isinstance(value, float):
            return str(value)
        elif isinstance(value, bytes):
            return 'X\'{:s}\''.format(value.hex())
        elif value is None:
            return 'null'
        else:
            value = str(value).replace('\"', '\\\"')
            return '"{:s}"'.format(value)


class ItemMysqlInsertPipeline(ItemBaseMysqlPipeline):
    sql_format = 'INSERT IGNORE INTO `{database}`.`{table}` ({fields}) ' \
                 'VALUES ({values})'

    async def process(self, item: Item) -> Item:
        fields = []
        values = []
        for k, v in self.to_dict(item).items():
            fields.append(k)
            values.append(self.convert_item_value(v))
        sql = self.sql_format.format(database=self.database, table=self.table,
                                     fields='`' + '`,`'.join(fields) + '`',
                                     values=','.join(values))
        await self.push_data(sql)
        return item


class ItemMysqlUpdatePipeline(ItemMysqlInsertPipeline):
    sql_format = 'UPDATE `{database}`.`{table}` SET {pairs} WHERE ' \
                 '`{primary_key}`={primary_value}'

    def __init__(self, *, primary_key: str,  **kwargs):
        super().__init__(**kwargs)
        self.primary_key = primary_key

    async def process(self, item: Item) -> Item:
        pairs = []
        primary_value = None

        for k, v in self.to_dict(item).items():
            if k == self.primary_key:
                primary_value = self.convert_item_value(v)
            else:
                pairs.append(
                    '`{:s}`={:s}'.format(k, self.convert_item_value(v)))

        if primary_value is not None:
            sql = self.sql_format.format(database=self.database,
                                         table=self.table,
                                         pairs=','.join(pairs),
                                         primary_key=self.primary_key,
                                         primary_value=primary_value)
            await self.push_data(sql)
        return item


class ItemMysqlInsertUpdatePipeline(ItemMysqlInsertPipeline):
    sql_format = 'INSERT INTO `{database}`.`{table}` ({fields}) VALUES ' \
                 '({values}) on duplicate key update {pairs}'

    def __init__(self, *, update_keys: List[str], **kwargs):
        super().__init__(**kwargs)
        self.update_keys = update_keys

    async def process(self, item: Item) -> Item:
        fields = []
        values = []
        pairs = []
        for k, v in self.to_dict(item).items():
            v = self.convert_item_value(v)
            fields.append(k)
            values.append(v)
            if k in self.update_keys:
                pairs.append('`{:s}`={:s}'.format(k, v))
        sql = self.sql_format.format(database=self.database, table=self.table,
                                     fields='`' + '`,`'.join(fields) + '`',
                                     values=','.join(values),
                                     pairs=','.join(pairs))
        await self.push_data(sql)
        return item


class ItemBaseEmailPipeline(Pipeline):
    def __init__(self, account: str, password: str, server: str, port: int,
                 recipients: List[str],
                 sender_name: str = 'AntNest.ItemEmailPipeline',
                 tls: bool = False, starttls: bool = False):
        super().__init__()
        self.account = account
        self.password = password
        self.server = server
        self.port = port
        self.recipients = recipients
        self.sender_name = sender_name
        self.starttls = starttls
        self.tls = tls

    async def create_smtp(self) -> aiosmtplib.SMTP:
        smtp = aiosmtplib.SMTP()
        if self.starttls:
            await smtp.connect(self.server, self.port, use_tls=False)
            await smtp.starttls()
        else:
            await smtp.connect(self.server, self.port, use_tls=self.tls)
        await smtp.login(self.account, self.password)
        return smtp

    async def send(self, smtp: aiosmtplib.SMTP, title: str, content: str,
                   attachments: Optional[List[IO]] = None):
        if attachments is None:
            msg = MIMEText(content)
        else:
            msg = MIMEMultipart()
            msg.attach(MIMEText(content))
            for f in attachments:
                att = MIMEText(f.read(), 'base64', 'utf-8')
                att["Content-Type"] = 'application/octet-stream'
                att["Content-Disposition"] = \
                    'attachment; filename="{:s}"'.format(f.name)
                msg.attach(att)
        msg['From'] = '{:s} <{:s}>'.format(self.sender_name, self.account)
        msg['To'] = '<' + '> <'.join(self.recipients) + '>'
        msg['Subject'] = title
        await smtp.send_message(msg)


class ItemEmailPipeline(ItemBaseEmailPipeline):
    def __init__(self, title, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.items = []
        self.title = title

    def process(self, thing: Item) -> Item:
        self.items.append(thing)
        return thing

    async def on_spider_close(self) -> None:
        smtp = await self.create_smtp()
        await self.send(smtp, self.title,
                        '\n'.join(item.__repr__() for item in self.items))
        smtp.close()


class ItemBaseRedisPipeline(Pipeline):
    def __init__(self, address: str, db: Optional[int] = None,
                 password: Optional[str] = None, encoding: str = 'utf-8',
                 minsize: int = 1, maxsize: int = 10,
                 ssl: Optional[bool] = None, timeout: Optional[float] = None):
        super().__init__()
        self.address = address
        self.db = db
        self.password = password
        self.encoding = encoding
        self.minsize = minsize
        self.maxsize = maxsize
        self.ssl = ssl
        self.timeout = timeout

    async def create_redis(self) -> aioredis.ConnectionsPool:
        return await aioredis.create_redis_pool(
            self.address, db=self.db, password=self.password,
            encoding=self.encoding, minsize=self.minsize,
            maxsize=self.maxsize, ssl=self.ssl, timeout=self.timeout)


__all__ = [var for var in vars().keys() if 'Pipeline' in var]



"""setting module for your project"""
import logging

logging.basicConfig(level=logging.INFO)

# your ant`s class modules or packages
ANT_PACKAGES = ['ants']

# custom setting, eg:
# MYSQL_HOST = '127.0.0.1'



import os
import sys
import asyncio
from unittest import mock

from yarl import URL
import pytest

from ant_nest import *
from ant_nest import CliAnt
from ant_nest.cli import *
from ant_nest import cli


def fake_response(content):
    res = Response('GET', URL('http://test.com'), writer=None,
                   continue100=None, timer=None,
                   request_info=None, traces=None,
                   loop=asyncio.get_event_loop(),
                   session=None)
    res._content = content
    res._body = content

    return res


def test_request():
    req = Request('GET', URL('http://test.com'))
    assert req.method == 'GET'
    req.__repr__()


def test_response():
    res = fake_response(b'1')
    assert res.get_text(encoding='utf-8') == '1'
    assert res.simple_text == '1'
    assert res.simple_json == 1

    res = fake_response(None)
    with pytest.raises(ValueError):
        res.get_text()
    res = fake_response(b'1')
    res.get_encoding = lambda: 'utf-8'
    res._get_encoding = lambda: 'utf-8'
    assert res.get_text() == '1'


def test_extract():
    with open('./tests/test.html', 'rb') as f:
        response = fake_response(f.read())
        response.get_text(encoding='utf-8')

    class Item:
        pass
    # extract item with xpath and regex
    item_extractor = ItemExtractor(Item)
    item_extractor.add_xpath('paragraph', '/html/body/div/p/text()')
    item_extractor.add_regex('title', '<title>([A-Z a-z]+)</title>',
                             item_extractor.extract_with_join_all)
    item = item_extractor.extract(response)
    assert item.paragraph == 'test'
    assert item.title == 'Test html'
    # some exception will be ignored
    item_extractor.add_regex('test', 'test(\d+)test')
    item = item_extractor.extract(response)
    assert getattr(item, 'test', None) is None
    # with exception for multiple result
    item_extractor.add_xpath('paragraph', '/html/head/title/text()')
    with pytest.raises(ItemExtractError):
        item = item_extractor.extract(response)
    # extract with jpath
    response = fake_response(b'{"a": {"b": {"c": 1}}, "d": null}')
    response.get_text(encoding='utf-8')
    item_extractor = ItemExtractor(Item)
    item_extractor.add_jpath('author', 'a.b.c')
    item_extractor.add_jpath('freedom', 'd')
    item = item_extractor.extract(response)
    assert item.author == 1
    assert item.freedom is None  # "None" obj can be extracted from json
    # extract single value with ValueError
    with pytest.raises(ValueError):
        ItemExtractor.extract_value('something else', 'test', 'test')
    # ItemNestExtractor tests
    with open('./tests/test.html', 'rb') as f:
        response = fake_response(f.read())
        response.get_text(encoding='utf-8')
    item_nest_extractor = ItemNestExtractor('xpath', '//div[@id="nest"]/div', Item)
    item_nest_extractor.add_xpath('xpath', './p/text()')
    item_nest_extractor.add_regex('regex', 'regex(\d+)</')
    temp = 1
    for item in item_nest_extractor.extract_items(response):
        assert item.xpath == str(temp)
        assert item.regex == str(temp)
        temp += 1

    with pytest.raises(NotImplementedError):
        item_nest_extractor.extract(response)
    # extract with wrappers
    with open('./tests/test.html', 'rb') as f:
        response = fake_response(f.read())
        response.get_text(encoding='utf-8')
    assert extract_value_by_xpath('/html/body/div/p/text()', response.html_element) == 'test'
    assert extract_value_by_xpath('/html/body/div/p/text()', response) == 'test'
    assert extract_value_by_xpath('//a/text()', '<a>test</a>', ignore_exception=False) == 'test'
    assert extract_value_by_jpath('a', {'a': 1}) == 1
    assert extract_value_by_jpath('a', '{"a": 1}') == 1
    assert extract_value_by_regex('(\d+)', 'I have 2 apples') == '2'
    assert extract_value_by_jpath('a', {'a': None}) is None
    assert extract_value_by_jpath('a', {}) is None  # default
    assert extract_value_by_jpath('a', {}, default=1) == 1
    with pytest.raises(Exception):
        extract_value_by_jpath('a', None, ignore_exception=False)


def test_cli_get_ants():
    ants = get_ants(['ant_nest', 'tests'])
    assert CliAnt is list(ants.values())[0]


def test_cli_shutdown():
    ant = CliAnt()
    ant.pool._queue.put_nowait(object())
    cli.shutdown_ant(ant)
    assert ant.pool._is_closed
    assert ant.pool._queue.qsize() == 0

    with pytest.raises(SystemExit):
        cli.shutdown_ant(ant)


def test_cli_run_ant():
    ant = CliAnt()
    asyncio.get_event_loop().run_until_complete(run_ant(ant))


def test_cli():

    with pytest.raises(SystemExit):
        cli.main(['-v'])

    with pytest.raises(SystemExit):  # no settings.py
        cli.main(['-l'])

    from ant_nest import _settings_example as settings
    # mock settings.py import
    sys.modules['settings'] = settings

    settings.ANT_PACKAGES = ['NoAnts']
    with pytest.raises(SystemExit):  # can`t import NoAnts
        cli.main(['-l'])

    settings.ANT_PACKAGES = ['ant_nest.things']
    with pytest.raises(SystemExit):  # no ants can be found
        cli.main(['-l'])

    settings.ANT_PACKAGES = ['tests']
    cli.main(['-l'])

    with pytest.raises(SystemExit):  # FakeAnt not exist
        cli.main(['-a' 'FakeAnt'])
        cli.main(['-l'])

    with pytest.raises(SystemExit), mock.patch('os.mkdir', lambda x: None):
        cli.main(['-c' '.'])

    cli.main(['-a' 'tests.test_things.CliAnt'])


def test_cli_open_browser():
    req = Request('GET', URL('http://test.com'))
    res = fake_response(b'')

    def open_browser_function(url):
        return True

    assert open_response_in_browser(
        res,
        _open_browser_function=open_browser_function)


def test_exception_filter():
    class FakeRecord:
        pass

    fr = ExceptionFilter()
    rc = FakeRecord()
    rc.exc_info = (ThingDropped, None, None)

    assert not fr.filter(rc)
    rc.exc_info = (OSError, None, None)
    assert fr.filter(rc)
